// Mega 1284 p


#define F_CPU 16000000UL                
#include "trtSettings.h"
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>

#include "trtkernel_1284.c"
// serial communication library
// Don't mess with the semaphores
#define FXNSTACKSIZE 350
#define SEM_RX_ISR_SIGNAL 1
#define SEM_STRING_DONE 2 // user hit <enter>
#define SEM_SHARED 3
#include "trtUart.h"
#include "trtUart.c"

// UART file descriptor
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
// input arguments to each thread, not actually used in this example
int args[3];

#define ZEROXG 512  //calibration for the lateral accel

void task1_pollADC(void);      		//execute at 5Hz (every 200ms)
void initialize(void); 				    //all the usual mcu stuff

volatile char Ain_high;
volatile char Ain_low;
volatile char Ain_low2; 
volatile int Ain; 
volatile int Ain_old = 865;

volatile int step_detected; //information from the wrist that is has calculated a footstep
volatile long prev_sum_accel = 0;
volatile long sum_accel = 0; //sum of acceleration information 
volatile int sum_accel_overflow;

 
//***********************      
// A to D conversion
void task1_pollADC(void)
{
  uint32_t rel,dead;
  unsigned long timestamp;
  Ain_old = 0;

  while (1) {
	
    //Wait for ADC done
    while (ADCSRA & (1<<ADSC) );
    //get the sample
    Ain_low = ADCL;
    Ain_low2 = ADCL;
    Ain_high = ADCH;
    Ain = (ADCH << 2) | (Ain_low2 >> 6);
    timestamp = trtCurrentTime(); //capture time of processing

    // AXIS 0 : Z AXIS 
    // {MUX1, MUX0} = 00, z axis data
    if (!(ADMUX & (1<<MUX1)) && !(ADMUX & (1<<MUX0))){
      //2.5V on MMA2260D is 0g
      //1.3V is +1g, 3.7V is -1g.

    	// if (Ain > 870){
      if (Ain > 868 || abs(Ain - Ain_old) > 30){
        //protect fprintf to prevent overlapping in packets -- can cause hangups
        trtWait(SEM_SHARED);
     		fprintf(stdout, "%d,%d,%lu,%ld\n", 0, Ain, timestamp, 0);
        trtSignal(SEM_SHARED);
        PORTC ^= 0x80; //toggle the transmit led
      }

      Ain_old = Ain;

    	ADMUX = (1<<ADLAR) | (1<<REFS0)| (0<< MUX1) | (1 << MUX0); //move onto channel 1
    }


    // AXIS 1: X AXIS
    // {MUX1, MUX0} = 01, x axis data
    else if (!(ADMUX & (1<<MUX1)) && (ADMUX & (1<<MUX0))) {
      //normalize around zero for 0g
      Ain = (Ain - ZEROXG);
      trtWait(SEM_SHARED);
      prev_sum_accel = sum_accel;
      //check for overflow
      if ( (sum_accel >0) && (LONG_MAX - Ain) < sum_accel){
        sum_accel_overflow++;
        sum_accel = sum_accel - (LONG_MAX - Ain);
      }
      //check for underflow
      else if ( (sum_accel < 0) && ((LONG_MIN - Ain) > sum_accel) ){
        sum_accel_overflow--;
        sum_accel = sum_accel - (LONG_MIN - Ain);
      }
      else { //todo there might be more cases where a positive sum_accel would get too negative
        sum_accel = sum_accel + Ain;
      }
      trtSignal(SEM_SHARED);

    	ADMUX = (1<<ADLAR) | (1<<REFS0)| (0<<MUX1) | (0<<MUX0); //move onto channel 0
    }

    //Kick off the conversion
    ADCSRA |= (1<<ADSC);

    //Sleep 
    rel = trtCurrentTime() + SECONDS2TICKS(.001);//0.0005); 
    dead = trtCurrentTime() + SECONDS2TICKS(.0011);//0.0006); 
    trtSleepUntil(rel, dead);
  }
} 

//*****************************
// Receive data from the Bluetooth
void task2_receive_step(void* args) {
  uint32_t rel,dead;
  unsigned long timestamp_prev = 0;
  unsigned long timestamp = 0; 
  unsigned long timestamp_diff = 0; 
  while (1) {

    fscanf(stdin, "%d", &step_detected);

    if (step_detected == 1){
      PORTC ^= 0x20;
      timestamp_prev = timestamp; 
      timestamp = trtCurrentTime();

      //Overflow case
      if (timestamp < timestamp_prev){
        timestamp_diff = timestamp + (ULONG_MAX - timestamp_prev);
      }
      else {
        timestamp_diff = timestamp - timestamp_prev;
      }

      //must protect fprintf statement to prevent overlapping in packets
      trtWait(SEM_SHARED);
      fprintf(stdout, "%d,%d,%lu,%ld\n", 1, sum_accel_overflow, timestamp_diff, sum_accel);
      sum_accel = 0; //reset the acceleration 
      sum_accel_overflow = 0;
      prev_sum_accel = 0;
      trtSignal(SEM_SHARED);
    }
    
  }
}

//***************************************************
// Notify us that we are progressing
void task3_keepalive(void* args){
  uint32_t rel, dead;
  while(1){
    PORTC ^= 0x08;

    rel = trtCurrentTime() + SECONDS2TICKS(.5);
    dead = trtCurrentTime() + SECONDS2TICKS(.55);
    trtSleepUntil(rel, dead);
  }
}

//**********************************************************
//Set it all up
void initialize(void)
{	
  //init the UART -- uart_init() is in uart.c
  trt_uart_init();
  stdout = stdin = stderr = &uart_str;
  
  DDRC |= 0xA8; //C.7 : transmit z-axis led
                //C.5 : received step from wrist
                //C.4 : keep alive LED

  //Set up the ADC
  //channel zero/ left adj /INTERNAL Aref=Vcc, MUx starts as 0
  //Left adjust means that the msb are the top 8 bits ADCH and lower 2 bits in ADCL
  ADMUX = (1<<ADLAR) | (1<<REFS0); 
  //enable and divide by 128, and kick off conversion
  ADCSRA=(1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0) | (1<<ADSC);
}

//Entry point for foot. 
int main(void){
  initialize();

  //start TRT
  trtInitKernel(150); //150 bytes for idle task stack

  // --- create semaphores ----------
  trtCreateSemaphore(SEM_RX_ISR_SIGNAL, 0) ; // uart receive ISR semaphore
  trtCreateSemaphore(SEM_STRING_DONE, 0) ; // task2_PID_control controls task1_PC_input rate
  trtCreateSemaphore(SEM_SHARED, 1) ; //protect fprintf!
  
  trtCreateTask(task1_pollADC, FXNSTACKSIZE, SECONDS2TICKS(0.0005), SECONDS2TICKS(0.0008), &(args[1])); //initially ADc takes.0002 to complete
  trtCreateTask(task2_receive_step, FXNSTACKSIZE, SECONDS2TICKS(0.05), SECONDS2TICKS(0.1), &(args[0]));
  trtCreateTask(task3_keepalive, FXNSTACKSIZE, SECONDS2TICKS(0.2), SECONDS2TICKS(.4), &(args[0]));

  while (1)
  {
    //loop de loop
  }
}

